""" Author: R. Le Gac

"""

import datetime
import json
import pprint
import traceback
import sys

from basesvc import BaseSvc
from gluon.http import HTTP
from gluon.restricted import RestrictedError
from helper import as_list

DBUI = 'Dbui'
PROC_KEY = '%s.%s'


class MyJsonEncoder(json.JSONEncoder):
    """Add date, datetime and time to the standard encoder.
    
    """
    def default(self, obj):
        
        if isinstance(obj, (datetime.date,
                            datetime.datetime,
                            datetime.time)):
            return obj.isoformat().replace('T', ' ')
        
        return json.JSONEncoder.default(self, obj)

        
    
class DirectSvc(BaseSvc):
    """Generic service to implement the Ext.Direct protocol on the server side
    Specification: http://www.sencha.com/products/extjs/extdirect/
    
    This implementation is based on the web2py service: 
    http://www.web2py.com/book/default/chapter/09
    It is a simplified version of the class Service in gluon.tools.
    
    To register functions and make it available remotely use decorator
    
        service = DirectSvc(globals())
        
        @service.register
        def myfunction(a, b):
    
    The same technique can be used with method of a class.
    
    On the client side the api definition can be retrieved using the
    url: /app/plugin_dbui/call/get_api

    On the client side the url to send requests is /app/plugin_dbui/call    
    Remote functions / methods can be call in the following way:
    Dbui.myfunction(a, b, callback), MyClass.mymethod(c, d, callback).
    
    In this implementation requests are JSON-Encoded raw HTTP
    and response are JSON encoded.
    
    """

    def __init__(self, environment):
        
        BaseSvc.__init__(self, environment)

        self.procedures = {}
        self.register(self.echo)


    def __call__(self):
        self.dbg('Start directSvc.call')
        request = self.environment['request']

        # no arguments in a Json-encoded raw HTPP request
        if len(request.args) == 0:
            return self.route()
        
        self.error(404, 'Object does not exist.')
        
    
    def echo(self, x):
        """Helper function returning the input arguments.
        Helpful to run echo tests.
        
        """
        return x
    

    def error(self, code, message):
        raise HTTP(code, message)
    
        
    def register(self, f):
        # a procedure is registered as 'action.method'
        # The action is equal to Dbui for a function
        # and to the name of the class for a method
        action = DBUI
        if 'im_class' in dir(f):
            action = f.im_class.__name__

        k = PROC_KEY % (action, f.__name__)
        self.procedures[k] = f
        
        self.dbg("Procedure %s registered" % k)
        return f

    
    def route(self):
        """Route the request to the appropriate methods, pass the 
        proper arguments and return the results.
        
        """
        self.dbg('Start directSvc.route')
        
        # decode transactions, a list of Json-encoded raw HTPP requests:
        # {action: "x", method: "y", data: [], type:"rpc", tid:z}
        request = self.environment['request']    
        transactions = as_list(json.loads(request.body.read()))
        
        # process each transactions and encode each result as a dict
        # {action: "x", method: "y", result: [], type:"rpc", tid:z}
        li = []
        for arg in transactions:
            di = dict(arg)
            self.dbg('transaction:', di)
            try:
                k = PROC_KEY % (arg['action'], arg['method'])
                if arg['data']:
                    di['result'] = self.procedures[k](*arg['data'])
                else:
                    di['result'] = self.procedures[k]()

            # catch exceptions and log them in the ticker system
            # inform the user that something went wrong in the server                
            except:
                r = RestrictedError(layer='directsvc.py',
                                    code='route',
                                    output='',
                                    environment=self.environment)
                r.log(request)
                self.error(500, 'Internal server error.')
            
            del di['data']
            li.append(di)
            
        self.dbg('End directSvc.call', li)
        return json.dumps(li, cls=MyJsonEncoder)
        
        