""" Author: R. Le Gac

"""
import copy

from fieldsmodifier import FieldsModifier
from gluon.storage import Storage
from helper import as_list
from modifier import Modifier


MODIFIER_FORMS = 'modifier_forms'


def configure_forms(db, **extjs):
    """Helper functions to apply to all grid widgets a set of 
    configuration options. Useful to set plugins for example.
    
    The keyword arguments contain the configuration options for
    the Ext.form.FormPanel.
    
    """
    for tablename in db.tables:
        fm = FormModifier(tablename)
        di = copy.deepcopy(extjs)
        fm.configure(**di)



class FormModifier(Modifier):
    """Helper tool to customise the form widget associated to a table.
    
    """
    
    def __init__(self, tablename):
        """Initialise the modifier persistent data.
        
        """
        Modifier.__init__(self, MODIFIER_FORMS, tablename)

        if 'field_sets' not in self.data:
            self.data.field_sets = []
            self.data.hidden_fields = []
            self.data.mapper = None

        self.tablename = tablename


    def link_comboboxes(self, **kwargs):
        """Link to Ext.form.ComboBox of the form in such a way that
        the value available for the second depend on the value selected
        in the first one. This first field is the master while the second one
        is the slave.
        
        This method modify deeply the field configuration transforming
        the Ext.form.comboBox into App.form.LinkedComboBox.
        
        The two comboboxes are linked to a common table and will display
        value of field in that table, including foreign fields.
        The table can be the one selected in the constructor or another
        one. The latter case is useful when working with selector.
         
        All keyword arguments of this method have to be defined:

                masterName
                    field name for the master combobox
                    
                slaveNames
                    field name for the master combobox.
                    Either a string or a list of string when several
                    slaves are linked to a master.
                    
                storeName
                    name of the store containing values
                    which will be displayed in the master and slave
                    combobox. Either mystoreStore for and Ext.data.ArrayStore
                    or tableStore for a store linked to a database table.
                    
                displayFieldMaster
                    Name of the field in the store shown in the master combo.
                    Either field for Ext.data.ArrayStore or TableField 
                    for a store linked to a database table.
                    
                valueFieldMaster
                    Name of the field in the store return by the master combo.
                    Either field for Ext.data.ArrayStore or TableField 
                    for a store linked to a database table.
                    
                displayFieldSlaves
                    Name of the field in the store display in the slave combo.
                    Either field for Ext.data.ArrayStore or TableField
                    for a store linked to a database table.
                    Either a string or a list of string when several
                    slaves are linked to a master.
                    
                valueFieldSlaves
                    Name of the field in the store return by the slave combo.
                    Either field for Ext.data.ArrayStore or TableField 
                    for a store linked to a database table.
                    Either a string or a list of string when several
                    slaves are linked to a master.
                    
                masterExtJS
                    configuration parameters dictionary for the master
                    Ext.form.Combobox. Might be useful to set the master
                    in readOnly mode for example
                    
                slavesExtJS
                    configuration parameters dictionary for the slave
                    Ext.form.Combobox. It can also be a list of dictionary
                    when running with multi-slaves. Might be useful to set 
                    the salve in readOnly mode for example.
                    
        """
        # Multi-slaves case
        t = ('slaveNames', 
             'displayFieldSlaves', 
             'valueFieldSlaves', 
             'slavesExtJS')
        
        for el in t:
            if el in kwargs:
                kwargs[el] = as_list(kwargs[el])

        # master
        extjs = {}
        if 'masterExtJS' in kwargs:
            extjs = kwargs['masterExtJS']
            
        fieldsModifier = FieldsModifier(self.tablename)
        fieldsModifier.configure_field(kwargs['masterName'],
                                       displayField=kwargs['displayFieldMaster'],
                                       role='master',
                                       store=kwargs['storeName'],
                                       valueField=kwargs['valueFieldMaster'],
                                       xtype='xlinkedcombobox',
                                       **extjs)
        
        # slaves
        for i in range(len(kwargs['slaveNames'])):

            extjs = {}
            if 'slavesExtJS' in kwargs:
                extjs = kwargs['slavesExtJS'][i]

            fieldsModifier.configure_field(kwargs['slaveNames'][i],
                                           displayField=kwargs['displayFieldSlaves'][i],
                                           mode='local',
                                           role='slave',
                                           store=kwargs['storeName'],
                                           valueField=kwargs['valueFieldSlaves'][i], 
                                           xtype='xlinkedcombobox',
                                           **extjs)


    def merge_fields(self, *fields, **kwargs):
        """Merge fields in a field set corresponding to 
        an Ext.form.FieldSet widget. 
        
        Each field is either a string identifying database field by its name
        or an instance of the modifier.Widget class (Spacer, ....).
        
        The keyword argument contains the configuration options of
        the Ext.form.FieldSet widget.
        
        """
        # by default the field occupies all the field set space
        extjs = {'defaults': {'anchor': '100%'}}
        extjs.update(kwargs)

        di = Storage(fields=fields, extjs= extjs)
        self.data.field_sets.append(di)


    def hide_fields(self, *fields):
        """Hide the list of fields in the form.
        Each field is identified by its database field name.
        
        """
        for el in fields:
            if el not in self.data.hidden_fields:
                self.data.hidden_fields.append(el)


    def set_mapper(self, func):
        """Link the mapper tool to the function func.
        
        A mapper tool is related to the organization of the field/fieldest
        in the form. By default the form layout is a column where fields are 
        pushed from the top to the bottom. By using the layout configuration 
        option it is possible to organize the field/fieldset in horizontal
        box, vertical box, columns or in table.
        
        This approach does not allow to map field/fieldset in panels
        contains in the form, like an Ext.TabPanel. The mapper tool allows to 
        handle this case. It is a function with one arguments, the list of 
        field/fieldset. The function return a dictionary with the structure of
        panels embedded in the form an the fields they contained.

        Example of mapper can be found in the mapper module.
        
        """
        self.data.mapper = func