""" A series of helper function

"""

import os
import re
import subprocess
import tempfile

from gluon import current
from gluon.dal import Table
from gluon.http import HTTP
from gluon.validators import (IS_DATE_IN_RANGE,
                              IS_DATETIME_IN_RANGE,
                              IS_DECIMAL_IN_RANGE,
                              IS_EMPTY_OR,
                              IS_FLOAT_IN_RANGE,
                              IS_IN_DB,
                              IS_IN_SET,
                              IS_INT_IN_RANGE,
                              IS_MATCH,
                              IS_LENGTH)

GIT = '/usr/bin/git'
STOREID = '%sStore'

# A field name can only contains [a_zA-Z0-9_] according to web2py
table_field = re.compile(r'[\w_]+\.[\w_]+') 
func_table_field = re.compile(r'([A-Z]+)\(([\w_]+)\.([\w_]+)\)')


def as_list(val):
    """Helper function returning val encapsulated into a list.
    If val is a list return val.  
        
    """
    if not isinstance(val, (list, tuple)):
         val = [val]
    return val


def decode_field(field):
    """Helper function returning (table, field) or (table, field, func). 

    """ 
    li = []
    for c in field:
        if c.isupper():
            li.append(c.lower())
        elif li:
            li[-1] += c
    return tuple(li)


def encode_field(*args):
    """Helper function returning TableField, TableFieldFunc.
    
    Note: comboBox don't decode properly any separator between 
    table and field like table+field. Therefore choice for the encoding
    are very limited.
    
    Note: don't use the method title since foo_faa becomes Foo_Faa and you
    want Foo_faa.
    
    """ 
    return ''.join([el[0].upper()+el[1:].lower() for el in args if len(el)])


def get_all_tables(dal):
    """Helper function to return a list of all gluon.dal.Table
    house by the DAL object. The list includes alias table.
    
    It is recommend to use standard iterator provided by 
    the DAL object. However by contruction, they ignore 
    alias table.
    
    """
    li = [dal[attr] for attr in dir(dal) if isinstance(dal[attr], Table)]
    return li

    
def get_create_id(table, **kwargs):
    """Helper function to find the id of a row identified by 
    a set of field value pairs.
    
    If the row does not exits it is created using fields, values pairs.
    Return the integer id of the row.

        table
            gluon.dal.Table
    
    """
    id = get_id(table, **kwargs)
    
    if id == None:
        id = table.insert(**kwargs)
        
    return id


def get_field_validators(field):
    """Helper function returning ExtJS configuration parameters 
    to handle database gluon.dal.Field validators on the client side.
    
    It mainly set value for minValue, maxValue, regular expression, ....
    
    Return a dictionary
    
    """
    cfg = {}
    
    validators = as_list(field.requires)
        
    for el in validators:
    
        vdt = el
        if isinstance(vdt, IS_EMPTY_OR):
            vdt = vdt.other
            
        # date and time validators
        if isinstance(vdt, IS_DATE_IN_RANGE):
            format = '%Y-%m-%d'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        elif isinstance(vdt, IS_DATETIME_IN_RANGE):
            format = '%Y-%m-%d %H:%M:%S'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        # number validators
        elif isinstance(vdt, (IS_DECIMAL_IN_RANGE, IS_FLOAT_IN_RANGE, IS_INT_IN_RANGE)):
            cfg['minValue'] = vdt.minimum
            cfg['maxValue'] = vdt.maximum
        
        # string validator
        elif isinstance(vdt, IS_MATCH):
            cfg["regex"] = vdt.regex.pattern
            cfg["regexText"] = vdt.error_message
        
        # others validator
        elif isinstance(vdt, IS_LENGTH):
            cfg["maxLength"] = vdt.maxsize
            cfg["minLength"] = vdt.minsize
            
    return cfg


def get_file_paths(path, ext=None, alpha=True):
    """Return a list of files paths according to the relative path:
    The relative path for a file or a directory is relative to the application.
    A path can be:
    
            None.
            The method return an empty list.
            
            A string pointing to one files.
            The method return a list with the local path of the file

            A list of string pointing to several files.
            The method return list of local paths for the selected files.
            
            A string pointing to a directory.
            The method return a list of files found in the directory.
           
            A list of string pointing to files and directory.
            The method return a list with local paths of the selected files 
            and those located in the selected directory.
    
    The return path is the client path which can be used in view.
    
    The keyword argument ext allows to filter files located in directory 
    by their extension: '.js', '.css'.
    
    If alpha is true the list of path is sort in alphabetical order.
    
    """
    li = []
    if path == None:
        return li
    
    server_path, client_path = get_reference_paths()
    
    for el in as_list(path):
        p = os.path.join(server_path, el)
        
        if os.path.isfile(p):
            if ext and (not el.endswith(ext)): continue
            li.append(os.path.join(client_path, el))
            
        elif os.path.isdir(p):
            for file in os.listdir(p):
                pfile = os.path.join(server_path, el, file)
                if os.path.isfile(pfile):
                    if ext and (not pfile.endswith(ext)): continue
                    li.append(os.path.join(client_path, el, file))
    
    if alpha:
        li.sort()
        
    return li


def get_foreign_field(field):
    """Helper function returning a tuple (k_tablename, k_fieldname, k_keyname)
    where:
        
        k_tablename is the name of the parent table
        
        k_fieldname is the name of the parent field
            table.field --> k_table.k_field
        
        k_keyname is the name of the primary key of the parent table
        requires by the left join. In most of the case it is id
            table.field == k_table.k_key
     
    Otherwise return None.
    
    """
    if not is_foreign_field(field):
        return None
    
    # foreign field validators IS_IN_BD contains all information 
    # on the parent key for the left join and on the parent field
    validators = as_list(field.requires)

    # Several syntax are allowed

    # IS_IN_DB is use.
    for vdt in validators:
        
        if isinstance(vdt, IS_IN_DB):
            # IS_IN_DB contains only the parent field. 
            # The key for the left join is the id key of the parent table
            if len(vdt.ks) == 1:
                return (vdt.ktable, vdt.ks[0], 'id')
            
            # IS_IN_DB contains the key for the join and the parent field 
            else:
                return (vdt.ktable, vdt.ks[0], vdt.ks[1])   
    
    # IS_IN_DB is not use. 
    # The parent key for the left join is the id key of the parent table
    # the parent field is the first field of the parent table which differs id
    if not validators:
        i = field.type.index(' ')
        ktablename = field.type[i:].strip()
        
        for kfield in field._db[ktablename]:
            if kfield.name != 'id':
                return (ktablename, kfield.name, 'id')
            

    return None


def get_id(table, **kwargs):
    """Helper function to find the id of a row identified by 
    a set of field value pairs.
    
    Return the integer id or None if the row does not exist.

        table
            gluon.dal.Table
            
    """
    query = None
    for k, v in kwargs.iteritems():
        el = table[k] == v

        if query:
            query = query & (el)
        else:
            query = el
   
    row =  table(query)
    if row:
        return row.id
           
    return None


def get_language():
    """Helper method returning the application language compliant with
    the ExtJS local file name.
    
    """
    lg = current.T.accepted_language
    
    # HTML tag for language: primary-dialect
    # Extract the primary language and ignore dialect
    m = re.match('([a-z]+)(\-([a-zA-Z]*))?', lg)
    if m:
        return m.group(1)

    return lg


def get_plugin_path(plugin_name):
    """Helper function returning the local path of the plugin name:
    plugin_extjs, plugin_mathjax, ...
    The local path defined with respect to the application directory.
   
    The function return the path of the first directory starting with
    plugin_name.
    
    Return None if the plugin is not found.
    
    """
    server_path, client_path = get_reference_paths()

    # look for the full name of the plugin directory
    p_static = os.path.join(server_path, 'static')
    for el in os.listdir(p_static):
        if os.path.isdir(os.path.join(p_static, el)):
            if el.startswith(plugin_name):
                return os.path.join(client_path, 'static', el)

    # Nothing found
    return None

    
def get_reference_paths():
    """Helper method returning a tuple with the server and the local paths.

    """
    server_path = os.path.join('applications', current.request.application)
    client_path = os.path.join(os.path.sep, current.request.application)
    
    return (server_path, client_path)


def get_set_field(field):
    """Helper function returning a list of the set data
    for the gluon.dal.Field.
    
    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return vdt.theset
        
    return []


def get_script_path(plugin):
    """Helper method returning the local path of the main script.
    the local path is defined with respect to the application directory.

    There is many way to define the main script via the url and/or via the
    configuration parameters of the plugin (application, base).
    
    If the script argument is found in the url, the main script is search in 
    script_dir, i.e static/scripts. the script name is given by the argument 
    value.
    
    Next in the list is the application script which might be defined in the 
    plugin configuration parameters. If no found the script associate to the
    plugin is used.
    
    """
    server_path, client_path = get_reference_paths()

    # script name specifies in the URL
    if "script" in current.request.vars:
        script = current.request.vars.script
        
        if not script.endswith(".js"):
            script = '%s.js' % script

        pdir = os.path.join(server_path, plugin.app_script_dir)
        
        if script not in os.listdir(pdir):
            raise HTTP(500, 'Request script "%s" does not exist !!!' % script)
        
        return os.path.join(client_path, plugin.app_script_dir, script)

    # default script for the application
    if plugin.app_script:
        return os.path.join(client_path, plugin.app_script)
        
    # default script for the plugin
    return os.path.join(client_path, plugin.dbui_script)


def get_store_id(name):
    """Heper function building the storeId from name
    
    """
    return STOREID % name


def get_versions():
    """Helper function returning the versions of the underlying libraries.
    
    """
    server_path, client_path = get_reference_paths()

    # plugin dbui
    fn = os.path.join(server_path, 'static', 'plugin_dbui', 'src', 'base.js')
    
    fi = open(fn, 'rb')
    s = fi.read()
    fi.close()
    
    m = re.match(".+App.version = '([\w._-]*)';.*", s, re.DOTALL)
    dbui = m.group(1)

    # plugin extjs
    extjs = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_extjs')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static', 
                                       'plugin_extjs', 
                                       'src', 
                                       'ext-core', 
                                       'src', 
                                       'core', 
                                       'Ext.js')
        
        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()
        
        m = re.match(r".+ version : '(\d+(\.\d+)*)',", s, re.DOTALL)
        if m:
            extjs = m.group(1)
    
    # plugin mathjax
    mathjax = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_mathjax')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static', 'plugin_mathjax', 'MathJax.js')
        
        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()
        
        m = re.match(r'.+MathJax.version="(\d+(\.\d+)*)";', s, re.DOTALL)
        if m:
            mathjax = m.group(1)

    # web2py
    li = [str(el) for el in current.request.env.web2py_version if isinstance(el, int)]
    web2py = '.'.join(li)
    
    # version of the application
    myapp = current.T("unknown")
    
    if os.path.exists(GIT):
        ref_path = os.getcwd()
        os.chdir(server_path)
    
        fi = tempfile.TemporaryFile()
        subprocess.call(["git", "describe", "--tags"], stdout=fi)
        fi.seek(0)
        myapp = fi.read().strip('\n')
        fi.close()
        
        os.chdir(ref_path)

    return dict(dbui=dbui, extjs=extjs, mathjax=mathjax, web2py=web2py, myapp=myapp)


def get_where_query(table):
    """Helper function build the where query in order to resolve foreign keys
    for the gluon.dal.Table table.
    
    Return None if there is no foreign key
    
    """
    db = table._db
    query = None

    for field in table:
        if not is_foreign_field(field):
            continue
        
        k_table, k_field, k_key = get_foreign_field(field)
        condition = field == db[k_table][k_key]
        
        if query:
            query = ((query) & (condition)) 
        else:
            query = condition
            
    return query

    
def is_foreign_field(field):
    """Helper function returning true if the gluon.dal.field 
    is a foreign key.
    
    """
    return field.type.startswith("reference")


def is_mathjax():
    """Helper function returning true is the mathjax plugin is installed.
    
    """
    return get_plugin_path('plugin_mathjax') != None


def is_set_field(field):
    """Helper function returning true if the gluon.dal.field value 
    should be in a set of values.
    
    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return True
        
    return False


def is_table_with_foreign_fields(table):
    """Helper function returning true if the table has at least
    a foreigne key.
    
    """
    for field in table:
        if is_foreign_field(field):
            return True
    return False


def rows_serializer(rows):
    """Helper function to serialize a gluon.dal.Rows.
    
    Handle the general case where field belongs to different
    table and where function can be applied.
    
    Field are encoded as TableField.
    
    Field with an applied function like COUNT(table.field) 
    are encoded as TableFieldCount
    
    Return a list of records
    A record is a dictionary {TableField: value,...}
    
    """
    # Map between column name and TableField
    map = {}
    for col in rows.colnames:
        if not table_field.match(col):
            func, table, field = func_table_field.match(col).groups()
            map[col] = encode_field(table, field, func)
        else:
            table, field = col.split(".")
            map[col] = encode_field(table, field)

    # serialize
    li = []
    for row in rows:
        di = {}
        for col in rows.colnames:
            if not table_field.match(col):
                di[map[col]] = row._extra[col]
            else:
                (table, field) = col.split(".")
                if table in row:
                    di[map[col]] = row[table][field]
                else:
                    di[map[col]] = row[field]
        li.append(di)

    return li