""" Base tools to build report

A database report relies on three main components:
    - A selector table and its form to allow the user to defined conditions.
      It can be build using a virtual database.
    - The Selector running in a controller to decode the user requirement
      and to select records in the database.
    - A view to render the selected records.

A powerful mechanism is also in place to render the report in different format
like HTML, CSV, .... It relies on the Selector class, on the IFRAME and 
on the view:

    selector = Selector(virtdb.myselector, extfield='format')
    iframe = selector.download()
    if iframe:
        return iframe

"""

from datetime import date, datetime, time
from gluon import current
from gluon.html import IFRAME, URL
from gluon.storage import Storage
from helper import (decode_field, 
                    get_foreign_field, 
                    is_foreign_field, 
                    is_table_with_foreign_fields)
                         

class Selector(Storage):
    """Basic tool to build a report. 
    
    Decode the data send by the selector widget. 
    Build the query for a given table of the database.
    Select Rows for a given table of the database.
    
    This class handle table and its foreign keys.
    The basic query handles the selector constraints
    and the inner join for foreign keys. Additional
    constraints can be added via the method add_query.

    All fields of the selector are attributes of this class.
    They can be accessed via myselector.myfield or myselector['myfield'].
    There type are those defined in the model (virtdb.myselector)
    
    A mechanism is provide to download the report in different format
    which can be selected by the user. It relies on the extfield and
    on the method download.
            
    """    
    def __init__(self, table, exclude_fields=(), extfield='format'):
        """
                table
                    the gluon.dal.Table containing the 
                    definition of the selector
                    
                exclude_fields
                    A tuple containing a list of selector field names
                    which will be ignored in query and select operations.
                    
                extfield
                    the name of the selector string defining the format
                    of the report. To be used in conjunction with the
                    download method.
                    
        """

        self._exclude_fields = exclude_fields
        self._ext_field = extfield
        self._extension = None
        self._extra_queries = []
        self._tablename = table._tablename
        
        # Decode the current request
        for key in current.request.vars:
            t = decode_field(key)
            
            if len(t) != 2 :
                continue
            
            fieldname = t[1]
            if fieldname == self._ext_field:
                self._extension = current.request.vars[key]

            # ignore field id
            elif fieldname == 'id':
                continue
            
            val = current.request.vars[key]
            
            if not val:
                self[fieldname] = val
                continue
            
            # convert the selector value according to field type
            fieldtype = table[fieldname].type

            if fieldtype == 'boolean':
                val = val == 'true' or val == 'True'
                
            elif fieldtype == 'date':
                ds = datetime.strptime(val, '%Y-%m-%dT%H:%M:%S')
                val = date(ds.year, ds.month, ds.day)
            
            elif fieldtype == 'datetime':
                val = datetime.strptime(val, '%Y-%m-%dT%H:%M:%S')
            
            elif fieldtype == 'double':
                val = float(val)
                
            elif fieldtype == 'integer':
                val = int(val)
                
            elif fieldtype == 'time':
                ds = datetime.strptime(val, '%H:%M:%S')
                val = time(ds.hour, ds.minute, ds.second)
                
            self[fieldname] = val


    def append_query(self, query):
        """Add constraints which will be used in the database query
        and database select operations.
            
            query
                gluon.dal.Query
        
        """
        self._extra_queries.append(query)
        
        
    def as_dict(self):
        """Return the selector fields and their values as a dictionary.
        
        """
        di = {}
        for field in self.fields():
            di[field] = self[field]
            
        return di
    
    
    def download(self):
        """The report can be download and receive as a file with a given format.
        this method initiate this process by return and IFRAME.
        The IFRAME is used by the browser to download the file.
        
        The file format is defined by one of the field of the list selector.
        By default is it 'format'. It can be changed via the argument extfield
        of the constructor.
        
        The IFRAME contain and URL. it it the same as the one defined in the
        current request (application/controller/function) but the extension 
        is defined by the format field.
        
        In this process all data conversion are performed by the view.
        
        """
        if  self._extension and self._extension != 'html':
    
            # remove the field use to select the file format
            # in order to avoid circular loop
            for key in current.request.vars:
                t = decode_field(key)
                if self._ext_field in t:
                    del current.request.vars[key]
                    break
    
            # The new URL is equal to the one defined in the current
            # request but with the proper extension
            url = URL(current.request.function, extension=self._extension, 
                                                vars=current.request.vars)
            
            # Return the IFRAME
            # the browser will request the file locate at the URL of the IFRAME
            # The IFRAME occupy all the space in the receiving container
            # Useful for embedded pdf application.
            return IFRAME(_frameborder=0,
                          _width="100%",
                          _height="100%",
                          _src=url)

        return None


    def fields(self):
        """Lists of fields defined in the selector.
        
        """
        li = []
        for k in self:
            if not k.startswith('_'):
                li.append(k)
        return li


    def query(self, table):
        """Build the database query for the database table
        including inner join for foreign keys, selector constraints
        and extra queries. The latter can be add via the append_query
        method.
        
        The extfield and fields in the exclude_fields tuple are 
        not take into account.
        
                table
                    gluon.dal.Table
                
        """
        db = table._db
        query = None
    
        # inner join for foreign keys
        if is_table_with_foreign_fields(table):
            for field in table:
                if is_foreign_field(field):
                    k_table, k_field, k_id = get_foreign_field(field)
                    q = field == db[k_table][k_id]
                    if query:
                        query = (query) & (q)
                    else:
                        query = q
                        
        # constraint from the selector
        for fieldname in self.fields():
            
            if not self[fieldname] or \
               fieldname in self._exclude_fields or \
               fieldname == self._ext_field:
                continue
            
            q = table[fieldname] == self[fieldname]
            
            if query:
                query = (query) & (q)
            else:
                query = q

        # append the extra queries
        for q in self._extra_queries:
            query = (query) & (q)
            
        return query


    def repr_url(self):
        """URL string corresponding to the current request.
        
        """
        vars = []
        
        tablename = self._tablename
        tablename = tablename[0].upper() + tablename[1:] 

        for (k, v) in current.request.vars.iteritems():
            if k.startswith(tablename) and len(v):
                vars.append("%s=%s" % (k, v))
    
        t = (current.request.env.http_host, 
             current.request.env.path_info, 
             '&'.join(vars))
        
        url = "http://%s%s?%s" % t
        return url


    def select(self, table, **kwargs):
        """Select the record of the table according to the selector constraints
        and to the extra queries.

        The extfield and fields in the exclude_fields tuple are 
        not take into account.
        
        It returns Rows which contains foreign tables columns. 
        
                table
                    gluon.dal.Table
                    
                kwargs
                    The keyword arguments are those of the 
                    gluon.dal.Set.select method: orderby, groupby, ....
        
        """
        db = table._db
        columns = [table.ALL]

        # columns for foreign tables
        if is_table_with_foreign_fields(table):
            for field in table:
                if is_foreign_field(field):
                    k_table, k_field, k_id = get_foreign_field(field)
                    columns.append(db[k_table].ALL)
                    
        # query handling selector and extra constraints
        query = self.query(table)
        
        # retrieve the selected records
        return db(query).select(*columns, **kwargs)